@@ -8,6 +8,7 @@ import zipfile
 
 from django.core.files.base import ContentFile
 from django.core.files.storage import default_storage
+from django.db.models import Q
 from django.template import Context, Template
 from django.utils import timezone
 from django.utils.module_loading import import_string
@@ -18,8 +19,20 @@ from six import string_types
 from web_fragments.fragment import Fragment
 from xblock.core import XBlock
 from xblock.completable import CompletableXBlockMixin
+from xblock.exceptions import JsonHandlerError
 from xblock.fields import Scope, String, Float, Boolean, Dict, DateTime, Integer
 
+try:
+    try:
+        from common.djangoapps.student.models import CourseEnrollment
+    except RuntimeError:
+        # Older Open edX releases have a different import path
+        from student.models import CourseEnrollment
+    from lms.djangoapps.courseware.models import StudentModule
+except ImportError:
+    CourseEnrollment = None
+    StudentModule = None
+
 
 # Make '_' a no-op so we can scrape strings
 def _(text):
@@ -30,6 +43,7 @@ logger = logging.getLogger(__name__)
 
 
 @XBlock.wants("settings")
+@XBlock.wants("user")
 class ScormXBlock(XBlock, CompletableXBlockMixin):
     """
     When a user uploads a Scorm package, the zip file is stored in:
@@ -129,6 +143,16 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
         default=False,
         scope=Scope.settings,
     )
+    popup_on_launch = Boolean(
+        display_name=_("Launch in pop-up window"),
+        help=_(
+            "Launch in pop-up window instead of embedding the SCORM content in "
+            "an iframe. Enable this for older packages that need to be run in "
+            "separate window."
+        ),
+        default=False,
+        scope=Scope.settings,
+    )
 
     has_author_view = True
 
@@ -137,6 +161,12 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
         template = Template(template_str)
         return template.render(Context(context))
 
+    def get_current_user_attr(self, attr: str):
+        return self.get_current_user().opt_attrs.get(attr)
+
+    def get_current_user(self):
+        return self.runtime.service(self, "user").get_current_user()
+
     @staticmethod
     def resource_string(path):
         """Handy helper for getting static resources from our kit."""
@@ -149,6 +179,7 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             context[
                 "message"
             ] = "Click 'Edit' to modify this module and upload a new SCORM package."
+        context["can_view_student_reports"] = True
         return self.student_view(context=context)
 
     def student_view(self, context=None):
@@ -156,18 +187,24 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             "index_page_url": self.index_page_url,
             "completion_status": self.lesson_status,
             "grade": self.get_grade(),
+            "can_view_student_reports": self.can_view_student_reports,
             "scorm_xblock": self,
         }
         student_context.update(context or {})
         template = self.render_template("static/html/scormxblock.html", student_context)
         frag = Fragment(template)
         frag.add_css(self.resource_string("static/css/scormxblock.css"))
+        frag.add_javascript(self.resource_string("static/js/src/scorm.js"))
         frag.add_javascript(self.resource_string("static/js/src/scormxblock.js"))
+        frag.add_javascript(self.resource_string("static/js/vendor/renderjson.js"))
         frag.initialize_js(
             "ScormXBlock",
             json_args={
                 "scorm_version": self.scorm_version,
                 "fullscreen_on_launch": self.fullscreen_on_launch,
+                "popup_on_launch": self.popup_on_launch,
+                "popup_width": self.width or 800,
+                "popup_height": self.height or 800,
                 "scorm_data": self.scorm_data,
             },
         )
@@ -183,6 +220,7 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             "field_width": self.fields["width"],
             "field_height": self.fields["height"],
             "field_fullscreen_on_launch": self.fields["fullscreen_on_launch"],
+            "field_popup_on_launch": self.fields["popup_on_launch"],
             "scorm_xblock": self,
         }
         studio_context.update(context or {})
@@ -207,6 +245,7 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
         self.has_score = request.params["has_score"] == "1"
         self.weight = parse_float(request.params["weight"], 1)
         self.fullscreen_on_launch = request.params["fullscreen_on_launch"] == "1"
+        self.popup_on_launch = request.params["popup_on_launch"] == "1"
         self.icon_class = "problem" if self.has_score else "video"
 
         response = {"result": "success", "errors": []}
@@ -229,6 +268,21 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
 
         return self.json_response(response)
 
+    @XBlock.handler
+    def popup_window(self, request, _suffix):
+        """
+        Standalone popup window
+        """
+        rendered = self.render_template(
+            "static/html/popup.html",
+            {
+                "index_page_url": self.index_page_url,
+                "width": self.width or 800,
+                "height": self.height or 800,
+            },
+        )
+        return Response(body=rendered)
+
     def clean_storage(self):
         if self.storage.exists(self.extract_folder_base_path):
             logger.info(
@@ -324,6 +378,10 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             return {"value": self.success_status}
         if name in ["cmi.core.score.raw", "cmi.score.raw"]:
             return {"value": self.lesson_score * 100}
+        if name in ["cmi.core.student_id", "cmi.learner_id"]:
+            return {"value": self.get_current_user_attr("edx-platform.user_id")}
+        if name in ["cmi.core.student_name", "cmi.learner_name"]:
+            return {"value": self.get_current_user_attr("edx-platform.username")}
         return {"value": self.scorm_data.get(name, "")}
 
     @XBlock.json_handler
@@ -332,15 +390,20 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
 
     @XBlock.json_handler
     def scorm_set_value(self, data, _suffix):
-        return self.set_value(data)
+        try:
+            return self.set_value(data)
+        except ValueError as e:
+            return JsonHandlerError(400, e.args[0]).get_response()
 
     def set_value(self, data):
         name = data.get("name")
+        value = data.get("value")
         completion_percent = None
         success_status = None
         completion_status = None
         lesson_score = None
 
+        self.scorm_data[name] = value
         if name == "cmi.core.lesson_status":
             lesson_status = data.get("value")
             if lesson_status in ["passed", "failed"]:
@@ -348,18 +411,13 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             elif lesson_status in ["completed", "incomplete"]:
                 completion_status = lesson_status
         elif name == "cmi.success_status":
-            success_status = data.get("value")
+            success_status = value
         elif name == "cmi.completion_status":
-            completion_status = data.get("value")
+            completion_status = value
         elif name in ["cmi.core.score.raw", "cmi.score.raw"] and self.has_score:
-            lesson_score = float(data.get("value", 0)) / 100.0
+            lesson_score = parse_validate_positive_float(value, name) / 100.0
         elif name == "cmi.progress_measure":
-            try:
-                completion_percent = float(data.get("value"))
-            except (ValueError, TypeError):
-                pass
-        else:
-            self.scorm_data[name] = data.get("value", "")
+            completion_percent = parse_validate_positive_float(value, name)
 
         context = {"result": "success"}
         if lesson_score is not None:
@@ -372,6 +430,8 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             context.update({"completion_status": completion_status})
         if success_status:
             self.success_status = success_status
+        if completion_status == "completed":
+            self.emit_completion(1)
         if success_status or completion_status == "completed":
             if self.has_score:
                 self.publish_grade()
@@ -397,7 +457,10 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
         """
         Utility method used to rescore a problem.
         """
-        self.lesson_score = score.raw_earned / self.weight
+        if self.has_score:
+            self.lesson_score = score.raw_earned
+            self.publish_grade()
+            self.emit_completion(1)
 
     def max_score(self):
         """
@@ -515,6 +578,81 @@ class ScormXBlock(XBlock, CompletableXBlockMixin):
             }
         return {}
 
+    @XBlock.handler
+    def scorm_search_students(self, data, _suffix):
+        """
+        Search enrolled students by username/email.
+        """
+        if not self.can_view_student_reports:
+            return Response(status=403)
+        query = data.params.get("id", "")
+        enrollments = (
+            CourseEnrollment.objects.filter(
+                is_active=True,
+                course=self.runtime.course_id,
+            )
+            .select_related("user")
+            .order_by("user__username")
+        )
+        if query:
+            enrollments = enrollments.filter(
+                Q(user__username__startswith=query) | Q(user__email__startswith=query)
+            )
+        # The format of each result is dictated by the autocomplete js library:
+        # https://github.com/dyve/jquery-autocomplete/blob/master/doc/jquery.autocomplete.txt
+        return self.json_response(
+            [
+                {
+                    "data": {"student_id": enrollment.user.id},
+                    "value": "{} ({})".format(
+                        enrollment.user.username, enrollment.user.email
+                    ),
+                }
+                for enrollment in enrollments[:20]
+            ]
+        )
+
+    @XBlock.handler
+    def scorm_get_student_state(self, data, _suffix):
+        if not self.can_view_student_reports:
+            return Response(status=403)
+        user_id = data.params.get("id")
+        try:
+            user_id = int(user_id)
+        except (TypeError, ValueError):
+            return Response(
+                body="Invalid 'id' parameter {}".format(user_id), status=400
+            )
+        try:
+            module = StudentModule.objects.filter(
+                course_id=self.runtime.course_id,
+                module_state_key=self.scope_ids.usage_id,
+                student__id=user_id,
+            ).get()
+        except StudentModule.DoesNotExist:
+            return Response(
+                body="No data found for student id={}".format(user_id),
+                status=404,
+            )
+        except StudentModule.MultipleObjectsReturned:
+            logger.error(
+                "Multiple StudentModule objects found for Scorm xblock: "
+                "course_id=%s module_state_key=%s student__id=%s",
+                self.runtime.course_id,
+                self.scope_ids.usage_id,
+                user_id,
+            )
+            raise
+        module_state = json.loads(module.state)
+        scorm_data = module_state.get("scorm_data", {})
+        return self.json_response(scorm_data)
+
+    @property
+    def can_view_student_reports(self):
+        if StudentModule is None:
+            return False
+        return getattr(self.runtime, "user_is_staff", False)
+
     @staticmethod
     def workbench_scenarios():
         """A canned scenario for display in the workbench."""
@@ -570,5 +708,17 @@ def parse_float(value, default):
         return default
 
 
+def parse_validate_positive_float(value, name):
+    try:
+        parsed = float(value)
+    except (TypeError, ValueError):
+        raise ValueError(
+            "Could not parse value of '{}' (must be float): {}".format(name, value)
+        )
+    if parsed < 0:
+        raise ValueError("Value of '{}' must not be negative: {}".format(name, value))
+    return parsed
+
+
 class ScormError(Exception):
     pass
